from typing import TYPE_CHECKING, Callable, Type, TypeVar

import jax
from jax import numpy as jnp

if TYPE_CHECKING:
    from .._base import MatrixLieGroup


T = TypeVar("T", bound="MatrixLieGroup")


def get_epsilon(dtype: jnp.dtype) -> float:
    """Helper for grabbing type-specific precision constants.

    Args:
        dtype: Datatype.

    Returns:
        Output float.
    """
    return {
        jnp.dtype("float32"): 1e-5,
        jnp.dtype("float64"): 1e-10,
    }[dtype]


def register_lie_group(
    *,
    matrix_dim: int,
    parameters_dim: int,
    tangent_dim: int,
    space_dim: int,
) -> Callable[[Type[T]], Type[T]]:
    """Decorator for registering Lie group dataclasses.

    - Sets static dimensionality attributes
    - Marks all methods for JIT compilation
    """

    def _wrap(cls: Type[T]) -> Type[T]:
        # Register dimensions as class attributes.
        cls.matrix_dim = matrix_dim
        cls.parameters_dim = parameters_dim
        cls.tangent_dim = tangent_dim
        cls.space_dim = space_dim

        # JIT all methods.
        for f in filter(
            lambda f: not f.startswith("_") and callable(getattr(cls, f)),
            dir(cls),
        ):
            setattr(cls, f, jax.jit(getattr(cls, f)))

        return cls

    return _wrap



from . import hints, manifold, utils
from ._base import MatrixLieGroup, SEBase, SOBase
from ._se2 import SE2
from ._se3 import SE3
from ._so2 import SO2
from ._so3 import SO3

__all__ = [
    "hints",
    "manifold",
    "utils",
    "MatrixLieGroup",
    "SOBase",
    "SEBase",
    "SE2",
    "SO2",
    "SE3",
    "SO3",
]



from typing import NamedTuple, Union

import numpy as onp
from jax import numpy as jnp

# Type aliases for JAX/Numpy arrays; primarily for function inputs.

Array = Union[onp.ndarray, jnp.ndarray]
"""Type alias for `Union[jnp.ndarray, onp.ndarray]`."""

Scalar = Union[float, Array]
"""Type alias for `Union[float, Array]`."""

Matrix = Array
"""Type alias for `Array`. Should not be instantiated.

Refers to a square matrix, typically with shape `(Group.matrix_dim, Group.matrix_dim)`.
For adjoints, shape should be `(Group.tangent_dim, Group.tangent_dim)`."""

Vector = Array
"""Type alias for `Array`. Should not be instantiated.

Refers to a general 1D array."""

TangentVector = Array
"""Type alias for `Array`. Should not be instantiated.

Refers to a 1D array with shape `(Group.tangent_dim,)`."""

# Type aliases for JAX arrays; primarily for function outputs.

ArrayJax = jnp.ndarray
"""Type alias for jnp.ndarray."""

ScalarJax = ArrayJax
"""Type alias for jnp.ndarray."""

MatrixJax = ArrayJax
"""Type alias for jnp.ndarray."""

VectorJax = ArrayJax
"""Type alias for jnp.ndarray."""

TangentVectorJax = ArrayJax
"""Type alias for jnp.ndarray."""


class RollPitchYaw(NamedTuple):
    """Tuple containing roll, pitch, and yaw Euler angles."""

    roll: Scalar
    pitch: Scalar
    yaw: Scalar


__all__ = [
    "Array",
    "Scalar",
    "Matrix",
    "Vector",
    "TangentVector",
    "ArrayJax",
    "ScalarJax",
    "MatrixJax",
    "VectorJax",
    "TangentVectorJax",
    "RollPitchYaw",
]
